package de.ozgcloud.nachrichten.postfach.osiv2.config;

import org.apache.hc.client5.http.auth.AuthScope;
import org.apache.hc.client5.http.auth.UsernamePasswordCredentials;
import org.apache.hc.client5.http.impl.auth.BasicCredentialsProvider;
import org.apache.hc.client5.http.impl.classic.HttpClientBuilder;
import org.apache.hc.core5.http.HttpHost;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.client.ClientHttpRequestFactory;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.http.converter.FormHttpMessageConverter;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.oauth2.client.AuthorizedClientServiceOAuth2AuthorizedClientManager;
import org.springframework.security.oauth2.client.InMemoryOAuth2AuthorizedClientService;
import org.springframework.security.oauth2.client.OAuth2AuthorizedClientProvider;
import org.springframework.security.oauth2.client.OAuth2AuthorizedClientProviderBuilder;
import org.springframework.security.oauth2.client.endpoint.RestClientClientCredentialsTokenResponseClient;
import org.springframework.security.oauth2.client.http.OAuth2ErrorResponseErrorHandler;
import org.springframework.security.oauth2.client.registration.ClientRegistrationRepository;
import org.springframework.security.oauth2.client.web.client.OAuth2ClientHttpRequestInterceptor;
import org.springframework.security.oauth2.core.http.converter.OAuth2AccessTokenResponseHttpMessageConverter;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestClient;

import de.ozgcloud.nachrichten.postfach.osiv2.gen.ApiClient;
import de.ozgcloud.nachrichten.postfach.osiv2.gen.api.MessageExchangeApi;
import lombok.RequiredArgsConstructor;

@Configuration
@EnableWebSecurity
@RequiredArgsConstructor
@ConditionalOnProperty(prefix = Osi2PostfachProperties.PREFIX, name = "enabled", havingValue = "true")
public class ApiClientConfiguration {

	private final Osi2PostfachProperties.ApiConfiguration apiConfiguration;
	private final Osi2PostfachProperties.ProxyConfiguration proxyConfiguration;

	@Bean
	MessageExchangeApi messageExchangeApi(ApiClient apiClient) {
		return new MessageExchangeApi(apiClient);
	}

	@Bean
	ApiClient apiClient(ClientRegistrationRepository clientRegistrations) {
		var apiClient = new ApiClient(restClient(clientRegistrations));
		apiClient.setBasePath(apiConfiguration.getUrl());
		return apiClient;
	}

	private RestClient restClient(ClientRegistrationRepository clientRegistrations) {
		var requestInterceptor = new OAuth2ClientHttpRequestInterceptor(authorizedClientManager(clientRegistrations));
		requestInterceptor.setClientRegistrationIdResolver(request -> "osi2");

		return defaultRestClientBuilder()
				.requestInterceptor(requestInterceptor)
				.build();
	}

	private RestClient.Builder defaultRestClientBuilder() {
		return RestClient.builder()
				.requestFactory(createProxyRequestFactory());
	}

	private ClientHttpRequestFactory createProxyRequestFactory() {
		var requestFactory = new HttpComponentsClientHttpRequestFactory();
		if (proxyConfiguration.isEnabled()) {
			requestFactory.setHttpClient(
					HttpClientBuilder.create()
							.setProxy(new HttpHost(proxyConfiguration.getHost(), proxyConfiguration.getPort()))
							.setDefaultCredentialsProvider(basicCredentialsProviderForProxy())
							.build()
			);
		}
		return requestFactory;
	}

	private BasicCredentialsProvider basicCredentialsProviderForProxy() {
		var credentialsProvider = new BasicCredentialsProvider();
		var username = proxyConfiguration.getUsername();
		var password = proxyConfiguration.getPassword();
		if (username != null && password != null) {
			credentialsProvider.setCredentials(new AuthScope(proxyConfiguration.getHost(), proxyConfiguration.getPort()),
					new UsernamePasswordCredentials(username, password.toCharArray()));
		}
		return credentialsProvider;
	}

	private AuthorizedClientServiceOAuth2AuthorizedClientManager authorizedClientManager(
			ClientRegistrationRepository clientRegistrations) {
		var clientService = new InMemoryOAuth2AuthorizedClientService(
				clientRegistrations);
		var authorizedClientManager = new AuthorizedClientServiceOAuth2AuthorizedClientManager(
				clientRegistrations, clientService);

		authorizedClientManager.setAuthorizedClientProvider(authorizedClientProvider());

		return authorizedClientManager;
	}

	private OAuth2AuthorizedClientProvider authorizedClientProvider() {
		return OAuth2AuthorizedClientProviderBuilder.builder()
				.clientCredentials(builder ->
						builder.accessTokenResponseClient(clientCredentialsTokenResponseClient())
				)
				.build();
	}

	private RestClientClientCredentialsTokenResponseClient clientCredentialsTokenResponseClient() {
		var client = new RestClientClientCredentialsTokenResponseClient();
		configureClientCredentialsRestClient(client);
		configureParametersForTokenRequests(client);
		return client;
	}

	private void configureClientCredentialsRestClient(RestClientClientCredentialsTokenResponseClient client) {
		client.setRestClient(defaultRestClientBuilder()
				.messageConverters(messageConverters -> {
					messageConverters.clear();
					messageConverters.add(new FormHttpMessageConverter());
					messageConverters.add(new OAuth2AccessTokenResponseHttpMessageConverter());
				})
				.defaultStatusHandler(new OAuth2ErrorResponseErrorHandler())
				.build());
	}

	private void configureParametersForTokenRequests(RestClientClientCredentialsTokenResponseClient client) {
		client.addParametersConverter(source -> {
			MultiValueMap<String, String> parameters = new LinkedMultiValueMap<>();
			// Pass a resource indicator parameter https://datatracker.ietf.org/doc/html/rfc8707
			parameters.add("resource", apiConfiguration.getResource());
			return parameters;
		});
	}

}
