package de.ozgcloud.nachrichten.postfach.osiv2;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.osiv2.factory.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osiv2.transfer.PostfachApiFacadeService;

class OsiPostfachRemoteServiceTest {

	@Spy
	@InjectMocks
	private OsiPostfachRemoteService osiPostfachRemoteService;

	@Mock
	private PostfachApiFacadeService postfachApiFacadeService;

	private final PostfachNachricht nachricht1 = PostfachNachrichtTestFactory.createBuilder()
			.subject("Nachricht 1")
			.build();
	private final PostfachNachricht nachricht2 = PostfachNachrichtTestFactory.createBuilder()
			.subject("Nachricht 2")
			.build();

	@DisplayName("send message")
	@Nested
	class TestSendMessage {

		@DisplayName("should call send message")
		@Test
		void shouldCallSendMessage() {
			postfachApiFacadeService.sendMessage(nachricht1);

			verify(postfachApiFacadeService).sendMessage(nachricht1);
		}

		@DisplayName("should throw osi postfach exception on runtime exception")
		@Test
		void shouldThrowOsiPostfachExceptionOnRuntimeException() {
			doThrow(new RuntimeException()).when(postfachApiFacadeService).sendMessage(nachricht1);

			assertThatThrownBy(() -> osiPostfachRemoteService.sendMessage(nachricht1))
					.isInstanceOf(OsiPostfachException.class);
		}
	}

	@DisplayName("get all messages")
	@Nested
	class TestGetAllMessages {

		@DisplayName("should return postfach messages")
		@Test
		void shouldReturnPostfachMessages() {
			when(postfachApiFacadeService.receiveMessages()).thenReturn(Stream.of(nachricht1, nachricht2));

			var result = osiPostfachRemoteService.getAllMessages();

			assertThat(result).containsExactly(nachricht1, nachricht2);
		}
	}

	@DisplayName("delete message")
	@Nested
	class TestDeleteMessage {

		@DisplayName("should call deleteMessage")
		@Test
		void shouldCallDeleteMessage() {
			postfachApiFacadeService.deleteMessage(UUID.randomUUID().toString());

			verify(postfachApiFacadeService).deleteMessage(any());
		}

		@DisplayName("should throw osi postfach exception on runtime exception")
		@Test
		void shouldThrowOsiPostfachExceptionOnRuntimeException() {
			doThrow(new RuntimeException()).when(postfachApiFacadeService).deleteMessage(UUID.randomUUID().toString());

			assertThatThrownBy(() -> osiPostfachRemoteService.deleteMessage(anyString())).isInstanceOf(OsiPostfachException.class);
		}
	}

	@DisplayName("get postfach type")
	@Nested
	class TestGetPostfachType {
		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = osiPostfachRemoteService.getPostfachType();

			assertThat(result).isEqualTo(OsiPostfachRemoteService.POSTFACH_TYPE_OSI);
		}
	}

	@DisplayName("is reply allowed")
	@Nested
	class TestIsReplyAllowed {
		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = osiPostfachRemoteService.isReplyAllowed();

			assertThat(result).isTrue();
		}
	}
}